@@ -23,19 +23,19 @@
 
 #include <assert.h>
 #include <ctype.h>
+#include <fcntl.h>
 
 #include "conffile.h"
 #include "dvdauthor.h"
 #include "readxml.h"
 #include "rgb.h"
 
-static const char RCSID[]="$Id: //depot/dvdauthor/src/dvdcli.c#52 $";
 
 #define RGB2YCrCb(R,G,B) ((((int)RGB2Y(R,G,B))<<16)|(((int)RGB2Cr(R,G,B))<<8)|(((int)RGB2Cb(R,G,B))))
 
 static int readdvdauthorxml(const char *xmlfile,char *fb);
 
-static int hadchapter=0,pauselen=0;
+static int hadchapter=0,pauselen=0,writeoutput=1;
 static char *chapters=0;
 
 static void parsevideoopts(struct pgcgroup *va,char *o)
@@ -173,14 +173,11 @@ static void parsechapters(char *o,struct source *src,int pauselen)
 static void readpalette(struct pgc *p,const char *fname)
 {
     int i,rgbf;
-    FILE *h;
+    struct vfile h;
 
-    if(fname[0] == '&' && isdigit(fname[1]) )
-        h=fdopen(atoi(&fname[1]),"rb");
-    else
-        h=fopen(fname,"rb");
+    h=varied_open(fname,O_RDONLY);
     
-    if( !h ) {
+    if( !h.h ) {
         fprintf(stderr,"ERR:  Cannot open palette file '%s'\n",fname);
         exit(1);
     }
@@ -189,7 +186,7 @@ static void readpalette(struct pgc *p,const char *fname)
     rgbf=( i>=4 && !strcasecmp(fname+i-4,".rgb") );
     for( i=0; i<16; i++ ) {
         int pcolor;
-        fscanf(h, "%x", &pcolor);
+        fscanf(h.h, "%x", &pcolor);
         if( rgbf ) {
             int r=(pcolor>>16)&255,
                 g=(pcolor>>8)&255,
@@ -205,7 +202,7 @@ static void readpalette(struct pgc *p,const char *fname)
 
     memset(groups,0,24);
     i=0;
-    while(fscanf(h,"%x",&b) == 1 && i<24 ) {
+    while(fscanf(h.h,"%x",&b) == 1 && i<24 ) {
         groups[i++]=b>>24;
         groups[i++]=b>>16;
         groups[i++]=b>>8;
@@ -214,7 +211,7 @@ static void readpalette(struct pgc *p,const char *fname)
             pgc_set_buttongroup(p,(i-8)/8,groups+(i-8));
     }
 #endif
-    fclose(h);
+    varied_close(h);
 }
 
 static void usage()
@@ -228,15 +225,17 @@ static void usage()
 #endif
 
     fprintf(stderr,
-            "syntax: dvdauthor [-o VTSBASE] [options] VOBFILE(s)\n"
+            "syntax: dvdauthor [-o VTSBASE | -n] [options] VOBFILE(s)\n"
             "\n\t-x XMLFILE where XMLFILE is a configuration file describing the\n"
             "\t    structure of the DVD to create.  If you use a config file, then you\n"
-            "\t    do not need to specify any other options, except -o.\n"
+            "\t    do not need to specify any other options, except -o and -n.\n"
+            "\n\t-n skips writing any files, for testing purposes.  MUST occur before any\n"
+            "\t    other options.\n"
             "\n\t" LONGOPT("--video=VOPTS or ") "-v VOPTS where VOPTS is a plus (+) separated list of\n"
             "\t    video options.  dvdauthor will try to infer any unspecified options.\n"
             "\t\tpal, ntsc, 4:3, 16:9, 720xfull, 720x576, 720x480, 704xfull,\n"
             "\t\t704x576, 704x480, 352xfull, 352x576, 352x480, 352xhalf,\n"
-            "\t\t352x288, 352x240, nopanscan, noletterbox.\n"
+            "\t\t352x288, 352x240, nopanscan, noletterbox, crop.\n"
             "\t    Default is ntsc, 4:3, 720xfull\n"
             "\n\t" LONGOPT("--audio=AOPTS or ") "-a AOPTS where AOPTS is a plus (+) separated list of\n"
             "\t    options for an audio track, with each track separated by a\n"
@@ -286,19 +285,23 @@ static void usage()
             "\t    inserted.\n"
             "\n\t" LONGOPT("--jumppad or ") "-j enables the creation of jumppads, which allow greater\n"
             "\t    flexibility in choosing jump/call desinations.\n"
+            "\n\t" LONGOPT("--allgprm or ") "-g enables the use of all 16 general purpose registers.\n"
             "\n\t" LONGOPT("--help or ") "-h displays this screen.\n"
         );
     exit(1);
 }
 
 #define MAINDEF                                                      \
+            if( !usedtocflag ) {                                     \
+                fprintf(stderr,"ERR:  Must first specify -t, -m, or -x.\n"); \
+                return 1;                                            \
+            }                                                        \
             if( istoc && curva ) {                                   \
                 fprintf(stderr,"ERR:  TOC cannot have titles\n");    \
                 return 1;                                            \
             }                                                        \
-            usedtocflag=1;                                           \
             if( !vc ) {                                              \
-                va[curva]=vc=pgcgroup_new(istoc+1-curva);             \
+                va[curva]=vc=pgcgroup_new(istoc+1-curva);            \
             }
 
 #define MAINDEFPGC                                                   \
@@ -343,6 +346,7 @@ int main(int argc,char **argv)
         {"entry",1,0,'e'},
         {"fpc",1,0,'F'},
         {"jumppad",0,0,'j'},
+        {"allgprm",0,0,'g'},
         {0,0,0,0}
     };
 #define GETOPTFUNC(x,y,z) getopt_long(x,y,"-" z,longopts,NULL)
@@ -360,7 +364,7 @@ int main(int argc,char **argv)
 
     while(1) {
         struct pgcgroup *vc=va[curva];
-        int c=GETOPTFUNC(argc,argv,"f:o:v:a:s:hc:Cp:Pmtb:Ti:e:x:");
+        int c=GETOPTFUNC(argc,argv,"f:o:v:a:s:hc:Cp:Pmtb:Ti:e:x:jgn");
         if( c == -1 )
             break;
         switch(c) {
@@ -375,10 +379,18 @@ int main(int argc,char **argv)
             }
             return readdvdauthorxml(optarg,fbase);
 
+        case 'n':
+            writeoutput=0;
+            break;
+
         case 'j':
             dvdauthor_enable_jumppad();
             break;
 
+        case 'g':
+            dvdauthor_enable_allgprm();
+            break;
+
         case 'T':
             if( usedtocflag ) {
                 fprintf(stderr,"ERR:  TOC (-T) option must come first because I am a lazy programmer.\n");
@@ -511,7 +523,7 @@ int main(int argc,char **argv)
     if( !va[1] && !istoc )
         va[1]=pgcgroup_new(0);
 
-    if( !fbase ) {
+    if( !fbase && writeoutput ) {
         fbase=readconfentry("WORKDIR");
         if( !fbase )
             usage();
@@ -520,24 +532,45 @@ int main(int argc,char **argv)
         fprintf(stderr,"ERR:  bad version of getopt; please precede all sources with '-f'\n");
         return 1;
     }
-    l=strlen(fbase);
-    if( l && fbase[l-1]=='/' )
-        fbase[l-1]=0;
+    if( fbase ) {
+        l=strlen(fbase);
+        if( l && fbase[l-1]=='/' )
+            fbase[l-1]=0;
+    }
 
     if( istoc )
         dvdauthor_vmgm_gen(fpc,mg,fbase);
     else
         dvdauthor_vts_gen(mg,va[1],fbase);
 
+    if( fpc )
+        pgc_free(fpc);
+    if( mg )
+        menugroup_free(mg);
+    if( va[1] )
+        pgcgroup_free(va[1]);
+
     return 0;
 }
 
+enum {
+    DA_BEGIN=0,
+    DA_ROOT,
+    DA_SET,
+    DA_PGCGROUP,
+    DA_PGC,
+    DA_VOB,
+    DA_SUBP,
+    DA_NOSUB
+};
+
 static struct pgcgroup *titles=0, *curgroup=0;
 static struct menugroup *mg=0, *vmgmmenus=0;
 static struct pgc *curpgc=0,*fpc=0;
 static struct source *curvob=0;
 static char *fbase=0,*buttonname=0;
-static int ismenuf=0,istoc=0,setvideo=0,setaudio=0,setsubpicture=0,hadtoc=0;
+static int ismenuf=0,istoc=0,setvideo=0,setaudio=0,setsubpicture=0,subpmode=DA_NOSUB,hadtoc=0,subpstreamid=-1;
+static char *subpstreammode=0;
 static int vobbasic,cell_chapter;
 static double cell_starttime,cell_endtime;
 static char menulang[3];
@@ -566,6 +599,14 @@ static void set_subpic_attr(int attr,char *s,int ch)
         pgcgroup_set_subpic_attr(titles,attr,s,ch);
 }
 
+static void set_subpic_stream(int ch,char *m,int id)
+{
+    if( ismenuf )
+        menugroup_set_subpic_stream(mg,ch,m,id);
+    else
+        pgcgroup_set_subpic_stream(titles,ch,m,id);
+}
+
 static int parse_pause(char *f)
 {
     if( !strcmp(f,"inf") )
@@ -590,9 +631,22 @@ static void dvdauthor_jumppad(char *s)
     }
 }
 
+static void dvdauthor_allgprm(char *s)
+{
+    int i=xml_ison(s);
+    if( i==1 )
+        dvdauthor_enable_allgprm();
+    else if( i==-1 ) {
+        fprintf(stderr,"ERR:  Unknown allgprm cmd '%s'\n",s);
+        exit(1);
+    }
+}
+
 static void getfbase()
 {
-    if( !fbase ) {
+    if( !writeoutput )
+        fbase=0;
+    else if( !fbase ) {
         fbase=readconfentry("WORKDIR");
         if( !fbase ) {
             fprintf(stderr,"ERR:  Must specify working directory\n");
@@ -603,10 +657,14 @@ static void getfbase()
 
 static void dvdauthor_end()
 {
-    if( fbase && hadtoc ) {
+    if( hadtoc ) {
         dvdauthor_vmgm_gen(fpc,vmgmmenus,fbase);
-        vmgmmenus=0;
+        if( fpc )
+            pgc_free(fpc);
+        if( vmgmmenus )
+            menugroup_free(vmgmmenus);
         fpc=0;
+        vmgmmenus=0;
     }
 }
 
@@ -619,10 +677,13 @@ static void titleset_start()
 static void titleset_end()
 {
     getfbase();
-    if( fbase ) {
+    if( !parser_err ) {
         if( !titles )
             titles=pgcgroup_new(0);
         dvdauthor_vts_gen(mg,titles,fbase);
+        if( mg )
+            menugroup_free(mg);
+        pgcgroup_free(titles);
         mg=0;
         titles=0;
     }
@@ -674,6 +735,7 @@ static void pgcgroup_start()
     setvideo=0;
     setaudio=-1;
     setsubpicture=-1;
+    subpmode=DA_PGCGROUP;
 }
 
 static void titles_start()
@@ -793,7 +855,21 @@ static void audio_channels(char *c)
     set_audio_attr(AUDIO_CHANNELS,ch,setaudio);
 }
 
-static void subattr_start()
+static void subattr_group_start()
+{
+    if( subpmode!=DA_PGCGROUP ) {
+        fprintf(stderr,"ERR:  Define all the subpictures before defining PGCs\n");
+        parser_err=1;
+        return;
+    }
+    setsubpicture++;
+    if( setsubpicture>=32 ) {
+        fprintf(stderr,"ERR:  Attempting to define too many subpicture streams for this PGC group\n");
+        parser_err=1;
+    }
+}
+
+static void subattr_pgc_start()
 {
     setsubpicture++;
     if( setsubpicture>=32 ) {
@@ -804,13 +880,53 @@ static void subattr_start()
 
 static void subattr_lang(char *c)
 {
-    set_subpic_attr(SPU_LANG,c,setsubpicture);
+    if( subpmode==DA_PGCGROUP )
+        set_subpic_attr(SPU_LANG,c,setsubpicture);
+    else {
+        fprintf(stderr,"ERR:  Cannot set subpicture language within a pgc; do it within titles or menus\n");
+        parser_err=1;
+    }
 }
 
+static void stream_start()
+{
+    subpstreamid=-1;
+    subpstreammode=0;
+}
+
+static void substream_id(char *c)
+{
+    subpstreamid=atoi(c);
+    if( subpstreamid<0 || subpstreamid>=32 ) {
+        fprintf(stderr,"ERR:  Subpicture stream id must be 0-31: '%s'.\n",c);
+        parser_err=1;
+    }
+}
+
+static void substream_mode(char *c)
+{
+    subpstreammode=strdup(c);
+}
+
+static void stream_end()
+{
+    if( subpstreamid==-1 || !subpstreammode ) {
+        fprintf(stderr,"ERR:  Must define the mode and id for the stream.\n");
+        parser_err=1;
+    } else if( subpmode==DA_PGCGROUP ) {
+        set_subpic_stream(setsubpicture,subpstreammode,subpstreamid);
+    } else
+        pgc_set_subpic_stream(curpgc,setsubpicture,subpstreammode,subpstreamid);
+    free(subpstreammode);
+}
+
+
 static void pgc_start()
 {
     curpgc=pgc_new();
     hadchapter=0;
+    setsubpicture=-1;
+    subpmode=DA_PGC;
 }
 
 static void pgc_entry(char *e)
@@ -834,6 +950,7 @@ static void pgc_end()
 {
     pgcgroup_add_pgc(curgroup,curpgc);
     curpgc=0;
+    subpmode=DA_NOSUB;
 }
 
 static void pre_start()
@@ -866,7 +983,9 @@ static void vob_start()
 
 static void vob_file(char *f)
 {
-    source_set_filename(curvob,utf8tolocal(f));
+    f=utf8tolocal(f);
+    source_set_filename(curvob,f);
+    free(f);
 }
 
 static void vob_chapters(char *c)
@@ -966,16 +1085,6 @@ static void button_end()
     if(buttonname) free(buttonname);
 }
 
-enum {
-    DA_BEGIN=0,
-    DA_ROOT,
-    DA_SET,
-    DA_PGCGROUP,
-    DA_PGC,
-    DA_VOB,
-    DA_NOSUB
-};
-
 static struct elemdesc elems[]={
     {"dvdauthor", DA_BEGIN,   DA_ROOT,    0,               dvdauthor_end},
     {"titleset",  DA_ROOT,    DA_SET,     titleset_start,  titleset_end},
@@ -985,19 +1094,22 @@ static struct elemdesc elems[]={
     {"menus",     DA_SET,     DA_PGCGROUP,menus_start,     menus_end},
     {"video",     DA_PGCGROUP,DA_NOSUB,   video_start,     0},
     {"audio",     DA_PGCGROUP,DA_NOSUB,   audio_start,     0},
-    {"subpicture",DA_PGCGROUP,DA_NOSUB,   subattr_start,   0},
+    {"subpicture",DA_PGCGROUP,DA_SUBP,    subattr_group_start, 0},
     {"pgc",       DA_PGCGROUP,DA_PGC,     pgc_start,       pgc_end},
     {"pre",       DA_PGC,     DA_NOSUB,   pre_start,       pre_end},
     {"post",      DA_PGC,     DA_NOSUB,   post_start,      post_end},
     {"button",    DA_PGC,     DA_NOSUB,   button_start,    button_end},
     {"vob",       DA_PGC,     DA_VOB,     vob_start,       vob_end},
+    {"subpicture",DA_PGC,     DA_SUBP,    subattr_pgc_start, 0},
     {"cell",      DA_VOB,     DA_NOSUB,   cell_start,      cell_end},
+    {"stream",    DA_SUBP,    DA_NOSUB,   stream_start,    stream_end},
     {0,0,0,0,0}
 };
 
 static struct elemattr attrs[]={
     {"dvdauthor","dest",dvdauthor_workdir},
     {"dvdauthor","jumppad",dvdauthor_jumppad},
+    {"dvdauthor","allgprm",dvdauthor_allgprm},
 
     {"menus","lang",menus_lang},
 
@@ -1018,13 +1130,18 @@ static struct elemattr attrs[]={
     {"video","resolution",video_resolution},
     {"video","widescreen",video_widescreen},
     {"video","caption",video_caption},
+
     {"audio","format",audio_format},
     {"audio","quant",audio_quant},
     {"audio","dolby",audio_dolby},
     {"audio","lang",audio_lang},
     {"audio","channels",audio_channels},
     {"audio","samplerate",audio_samplerate},
+
     {"subpicture","lang",subattr_lang},
+    {"stream","mode",substream_mode},
+    {"stream","id",substream_id},
+
     {"pgc","entry",pgc_entry},
     {"pgc","palette",pgc_palette},
     {"pgc","pause",pgc_pause},
