@@ -20,15 +20,27 @@
 #include "config.h"
 
 #include "compat.h"
+#include <assert.h>
 
 #include "dvdauthor.h"
 #include "da-internal.h"
 #include "dvdvm.h"
 
-static const char RCSID[]="$Id: //depot/dvdauthor/src/dvdcompile.c#15 $";
 
 struct vm_statement *dvd_vm_parsed_cmd;
 
+#define MAXLABELS 200
+#define MAXGOTOS 200
+
+struct dvdlabel {
+    char *lname;
+    unsigned char *code;
+};
+
+static struct dvdlabel labels[MAXLABELS];
+static struct dvdlabel gotos[MAXGOTOS];
+static int numlabels=0, numgotos=0;
+
 static int negatecompare(int compareop)
 {
     return compareop^1^((compareop&4)>>1);
@@ -56,14 +68,24 @@ static int compile_usesreg(struct vm_statement *cs,int target)
 
 static int nexttarget(int t)
 {
-    if( t<13 )
-        return 13;
-    t++;
-    if( t==16 ) {
-        fprintf(stderr,"ERR:  Expression is too complicated, ran out of registers\n");
-        exit(1);
+    if( !allowallreg ) {
+        if( t<13 )
+            return 13;
+        t++;
+        if( t<16 )
+            return t;
     }
-    return t;
+    fprintf(stderr,"ERR:  Expression is too complicated, ran out of registers\n");
+    exit(1);
+}
+
+// like nexttarget, but takes a VM_VAL argument
+static int nextval(int t)
+{
+    if( t<-128 )
+        return nexttarget(t+256)-256;
+    else
+        return nexttarget(-1)-256;
 }
 
 static unsigned char *compileop(unsigned char *buf,int target,int op,int val)
@@ -77,6 +99,7 @@ static unsigned char *compileop(unsigned char *buf,int target,int op,int val)
     case VM_MUL: buf[0]|=5; break;
     case VM_DIV: buf[0]|=6; break;
     case VM_MOD: buf[0]|=7; break;
+    case VM_RND: buf[0]|=8; break;
     case VM_AND: buf[0]|=9; break;
     case VM_OR:  buf[0]|=10; break;
     case VM_XOR: buf[0]|=11; break;
@@ -137,17 +160,28 @@ static unsigned char *compileexpr(unsigned char *buf,int target,struct vm_statem
             }
     }
 
+    // special case -- rnd where the parameter is a reg or value
+    if( cs->op == VM_RND && cs->param->op == VM_VAL ) {
+        assert(cs->param->next==0);
+        return compileop(buf,target,cs->op,cs->param->i1);
+    }
+
     buf=compileexpr(buf,target,cs->param);
-    for( v=cs->param->next; v; v=v->next ) {
-        if( v->op==VM_VAL && !issprmval(v))
-            buf=compileop(buf,target,cs->op,v->i1);
-        else {
-            int t2=nexttarget(target);
-            buf=compileexpr(buf,t2,v);
-            buf=compileop(buf,target,cs->op,t2-256);
-            if( t2==15 ) {
-                write8(buf,0x71,0x00,0x00,0x0f,0x00,0x00,0x00,0x00);
-                buf+=8;
+    if( cs->op == VM_RND ) {
+        assert(cs->param->next==0);
+        return compileop(buf,target,cs->op,target-256);
+    } else {
+        for( v=cs->param->next; v; v=v->next ) {
+            if( v->op==VM_VAL && !issprmval(v))
+                buf=compileop(buf,target,cs->op,v->i1);
+            else {
+                int t2=nexttarget(target);
+                buf=compileexpr(buf,t2,v);
+                buf=compileop(buf,target,cs->op,t2-256);
+                if( t2==15 ) {
+                    write8(buf,0x71,0x00,0x00,0x0f,0x00,0x00,0x00,0x00);
+                    buf+=8;
+                }
             }
         }
     }
@@ -170,17 +204,14 @@ static unsigned char *compilebool(unsigned char *obuf,unsigned char *buf,struct
         if( cs->param->op==VM_VAL )
             r1=cs->param->i1;
         else {
-            r1=13-256;
-            buf=compileexpr(buf,r1,cs->param);
+            r1=nextval(0);
+            buf=compileexpr(buf,r1&15,cs->param);
         }
         if( cs->param->next->op==VM_VAL && (r1<0 || cs->param->next->i1<0) )
             r2=cs->param->next->i1;
         else {
-            if( r1==13-256 )
-                r2=14-256;
-            else
-                r2=13-256;
-            buf=compileexpr(buf,r2,cs->param->next);
+            r2=nextval(r1);
+            buf=compileexpr(buf,r2&15,cs->param->next);
         }
         if( r1>=0 ) {
             int t;
@@ -241,7 +272,8 @@ static unsigned char *compilebool(unsigned char *obuf,unsigned char *buf,struct
     return buf;
 }
 
-static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct workset *ws,int thispgc,struct vm_statement *cs,int ismenu,int prepost)
+// NOTE: curgroup is passed separately from curpgc, because in FPC, curpgc==NULL, but curgroup!=NULL
+static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,const struct workset *ws,const struct pgcgroup *curgroup,const struct pgc *curpgc,const struct vm_statement *cs,int ismenu)
 {
     int lastif=0;
 
@@ -262,37 +294,75 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
             case 9:
             case 10:
             case 11:
-            case 12: // set GPRM
+            case 12:
+            case 13:
+            case 14:
+            case 15: // set GPRM
                 buf=compileexpr(buf,cs->i1,cs->param);
                 break;
 
+            case 32+0:
+            case 32+1:
+            case 32+2:
+            case 32+3:
+            case 32+4:
+            case 32+5:
+            case 32+6:
+            case 32+7:
+            case 32+8:
+            case 32+9:
+            case 32+10:
+            case 32+11:
+            case 32+12:
+            case 32+13:
+            case 32+14:
+            case 32+15: // set GPRM
+                if( cs->param->op==VM_VAL ) { // we can set counters to system registers
+                    int v=cs->param->i1;
+                    if( v<0 )
+                        write8(buf,0x43,0x00,0x00,v,0x00,0x80|(cs->i1-32),0x00,0x00);
+                    else
+                        write8(buf,0x53,0x00,v/256,v,0x00,0x80|(cs->i1-32),0x00,0x00);
+                    buf+=8;
+                } else {
+                    int r=nexttarget(0);
+                    buf=compileexpr(buf,r,cs->param);
+                    write8(buf,0x43,0x00,0x00,r,0x00,0x80|(cs->i1-32),0x00,0x00);
+                    buf+=8;
+                }
+                break;
+
             case 128+1: // audio
             case 128+2: // subtitle
             case 128+3: // angle
-                if( cs->param->op==VM_VAL && !( cs->i2>=-128 && cs->i2<0 )) {
+                if( cs->param->op==VM_VAL && !( cs->param->i1>=-128 && cs->param->i1<0 )) {
                     int v=cs->param->i1;
                     write8(buf,v<0?0x41:0x51,0x00,0x00,0x00,0x00,0x00,0x00,0x00);
                     buf[(cs->i1-128)+2]=128|v; // doesn't matter whether this is a register or a value
                     buf+=8;
                 } else {
-                    buf=compileexpr(buf,13,cs->param);
+                    int r=nexttarget(0);
+                    buf=compileexpr(buf,r,cs->param);
                     write8(buf,0x41,0x00,0x00,0x00,0x00,0x00,0x00,0x00);
-                    buf[(cs->i1-128)+2]=128|13;
+                    buf[(cs->i1-128)+2]=128|r;
                     buf+=8;
                 }
                 break;
                 
             case 128+8: // button
-                if( cs->param->op==VM_VAL && !( cs->i2>=-128 && cs->i2<0 )) {
+                if( cs->param->op==VM_VAL && !( cs->param->i1>=-128 && cs->param->i1<0 )) {
                     int v=cs->param->i1;
+                    if( v>0 && (v&1023)!=0 )
+                        fprintf(stderr,"WARN: Button value is %d, but it should be a multiple of 1024\nWARN: (button #1=1024, #2=2048, etc)\n",v);
                     if( v<0 )
                         write8(buf,0x46,0x00,0x00,0x00,0x00,v,0x00,0x00);
                     else
                         write8(buf,0x56,0x00,0x00,0x00,v/256,v,0x00,0x00);
                     buf+=8;
                 } else {
-                    buf=compileexpr(buf,13,cs->param);
-                    write8(buf,0x46,0x00,0x00,0x00,0x00,13,0x00,0x00);
+                    int r=nexttarget(0);
+                    buf=compileexpr(buf,r,cs->param);
+                    write8(buf,0x46,0x00,0x00,0x00,0x00,r,0x00,0x00);
                     buf+=8;
                 }
                 break;
@@ -308,9 +378,9 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
             while(1) {
                 unsigned char *lp,*ib,*e;
 
-                lp=compilecs(obuf,iftrue,ws,thispgc,cs->param->next->param,ismenu,prepost);
+                lp=compilecs(obuf,iftrue,ws,curgroup,curpgc,cs->param->next->param,ismenu);
                 if( cs->param->next->next ) {
-                    e=compilecs(obuf,lp+8,ws,thispgc,cs->param->next->next,ismenu,prepost);
+                    e=compilecs(obuf,lp+8,ws,curgroup,curpgc,cs->param->next->next,ismenu);
                     write8(lp,0x00,0x01,0x00,0x00,0x00,0x00,0x00,(e-obuf)/8+1);
                     lp+=8;
                 } else
@@ -324,10 +394,47 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
                 end=e;
             }
             buf=end;
-            lastif=1;
+            lastif=1; // make sure reference statement is generated
             break;
         }
 
+        case VM_LABEL:
+        {
+            int i;
+
+            for( i=0; i<numlabels; i++ )
+                if( !strcasecmp(labels[i].lname,cs->s1) ) {
+                    fprintf(stderr,"ERR:  Duplicate label '%s'\n",cs->s1);
+                    return 0;
+                }
+            if( numlabels == MAXLABELS ) {
+                fprintf(stderr,"ERR:  Too many labels\n");
+                return 0;
+            }
+            labels[numlabels].lname=cs->s1;
+            labels[numlabels].code=buf;
+            numlabels++;
+            lastif=1; // make sure reference statement is generated
+            break;
+        }
+            
+        case VM_GOTO:
+            if( numgotos == MAXGOTOS ) {
+                fprintf(stderr,"ERR:  Too many gotos\n");
+                return 0;
+            }
+            gotos[numgotos].lname=cs->s1;
+            gotos[numgotos].code=buf;
+            numgotos++;
+            write8(buf,0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00);
+            buf+=8;
+            break;
+            
+        case VM_BREAK:
+            write8(buf,0x00,0x02,0x00,0x00,0x00,0x00,0x00,0x00);
+            buf+=8;
+            break;
+
         case VM_EXIT:
             write8(buf,0x30,0x01,0x00,0x00,0x00,0x00,0x00,0x00);
             buf+=8;
@@ -339,7 +446,11 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
             break;
 
         case VM_JUMP:
-            if( cs->i1==1 && ismenu==2 ) {
+        {
+            int i1=cs->i1;
+            int i2=cs->i2;
+
+            if( i1==1 && ismenu==2 ) {
                 //	VMGM	VMGM	NOPGC	NOCH
                 //	VMGM	VMGM	NOPGC	CHXX
                 //	VMGM	VMGM	MPGC	NOCH
@@ -348,9 +459,9 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
                 //	VMGM	VMGM	MEPGC	CHXX
                 //	VMGM	VMGM	TPGC	NOCH
                 //	VMGM	VMGM	TPGC	CHXX
-                cs->i1=0;
+                i1=0;
             }
-            if( ((cs->i2>0 && cs->i2<128) || (cs->i2==0 && cs->i1==1)) && !ismenu ) {
+            if( ((i2>0 && i2<128) || (i2==0 && i1==1)) && !ismenu ) {
                 //	VTS	NONE	MPGC	NOCH
                 //	VTS	VMGM	MPGC	NOCH
                 //	VTS	TS	MPGC	NOCH
@@ -369,7 +480,7 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
                 fprintf(stderr,"ERR:  Cannot jump to a menu from a title, use 'call' instead\n");
                 return 0;
             }
-            if( cs->i2>0 && cs->i2<128 && cs->i3 && ismenu ) {
+            if( i2>0 && i2<128 && cs->i3 && ismenu ) {
                 //	VMGM	NONE	MPGC	CHXX
                 //	VMGM	TS	MPGC	CHXX
                 //	VMGM	NONE	MEPGC	CHXX
@@ -383,7 +494,7 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
                 fprintf(stderr,"ERR:  Cannot specify chapter when jumping to another menu\n");
                 return 0;
             }
-            if( cs->i1 && !cs->i2 ) {
+            if( i1 && !i2 ) {
                 //	VTSM	VMGM	NOPGC	CHXX
                 //	VTS	TS	NOPGC	CHXX
                 //	VTSM	TS	NOPGC	CHXX
@@ -394,14 +505,14 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
                 fprintf(stderr,"ERR:  Cannot omit menu/title if specifying vmgm/titleset\n");
                 return 0;
             }
-            if( !cs->i1 && !cs->i2 && !cs->i3 ) {
+            if( !i1 && !i2 && !cs->i3 ) {
                 //	VTS	NONE	NOPGC	NOCH
                 //	VTSM	NONE	NOPGC	NOCH
                 //	VMGM	NONE	NOPGC	NOCH
                 fprintf(stderr,"ERR:  Nop jump statement\n");
                 return 0;
             }
-            if( cs->i2==121 && (cs->i1>=2 || (cs->i1==0 && ismenu!=2)) ) {
+            if( i2==121 && (i1>=2 || (i1==0 && ismenu!=2)) ) {
                 fprintf(stderr,"ERR:  VMGM must be specified with FPC\n");
                 return 0;
             }
@@ -409,15 +520,15 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
 
 
             // *** ACTUAL COMPILING
-            if( cs->i1>=2 && cs->i2>=120 && cs->i2<128 ) {
+            if( i1>=2 && i2>=120 && i2<128 ) {
                 //	VTSM	TS	MEPGC	NOCH
                 //	VMGM	TS	MEPGC	NOCH
-                if( cs->i2==120 )
-                    cs->i2=123;
-                write8(buf,0x30,0x06,0x00,0x01,cs->i1-1,0x80+(cs->i2-120),0x00,0x00); buf+=8;
-            } else if( cs->i1>=2 ||
-                       ( cs->i1==1 && cs->i2>=128 ) ||
-                       ( ismenu==2 && cs->i2>=128 && cs->i3 )) {
+                if( i2==120 )
+                    i2=123;
+                write8(buf,0x30,0x06,0x00,0x01,i1-1,0x80+(i2-120),0x00,0x00); buf+=8; // JumpSS VTSM vts 1 menu
+            } else if( i1>=2 ||
+                       ( i1==1 && i2>=128 ) ||
+                       ( ismenu==2 && i2>=128 && cs->i3 )) {
                 //	VMGM	TS	TPGC	CHXX
                 //	VTSM	TS	MPGC	NOCH
                 //	VMGM	TS	MPGC	NOCH
@@ -432,36 +543,35 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
                 //	VTSM	VMGM	TPGC	CHXX
                 //	VMGM	NONE	TPGC	CHXX
                 if( jumppad ) {
-                    if( !cs->i1 )
-                        cs->i1=1;
-                    write8(buf,0x71,0x00,0x00,0x0F,cs->i2,cs->i1,0x00,0x00); buf+=8;
+                    if( !i1 )
+                        i1=1;
+                    write8(buf,0x71,0x00,0x00,0x0F,i2,i1,0x00,0x00); buf+=8;
                     write8(buf,0x71,0x00,0x00,0x0E,0x00,cs->i3,0x00,0x00); buf+=8;
                     write8(buf,0x30,ismenu?0x06:0x08,0x00,0x00,0x00,0x42,0x00,0x00); buf+=8;
                 } else {
                     fprintf(stderr,"ERR:  That form of jumping is not allowed\n");
                     return 0;
                 }
-            } else if( cs->i1==1 || cs->i2==121 ) {
+            } else if( i1==1 || i2==121 ) {
                 //	VTSM	VMGM	NOPGC	NOCH
                 //	VTSM	VMGM	MPGC	NOCH
                 //	VTSM	VMGM	MEPGC	NOCH
                 // cannot error check jumps to the vmgm menu
-                if( !cs->i2 || cs->i2==120 )
-                    cs->i2=122;
-                if( cs->i2<120 )
-                    write8(buf,0x30,0x06,0x00,cs->i2,0x00,0xC0,0x00,0x00);
+                if( !i2 || i2==120 )
+                    i2=122;
+                if( i2<120 )
+                    write8(buf,0x30,0x06,0x00,i2,0x00,0xC0,0x00,0x00); // JumpSS VMGM pgcn
                 else
-                    write8(buf,0x30,0x06,0x00,0x00,0x00,cs->i2==121?0:(0x40+cs->i2-120),0x00,0x00);
+                    write8(buf,0x30,0x06,0x00,0x00,0x00,i2==121?0:(0x40+i2-120),0x00,0x00); // JumpSS FP or JumpSS VMGM menu
                 buf+=8;
-            } else if( !cs->i1 && !cs->i2 && cs->i3 ) {
+            } else if( !i1 && !i2 && cs->i3 ) {
                 int numc;
-                struct pgc *p;
                 char *des;
 
                 //	VTS	NONE	NOPGC	CHXX
                 //	VTSM	NONE	NOPGC	CHXX
                 //	VMGM	NONE	NOPGC	CHXX
-                if( thispgc<0 ) {
+                if( curpgc==0 ) {
                     fprintf(stderr,"ERR:  Cannot jump to a chapter from a FPC\n");
                     return 0;
                 }
@@ -469,48 +579,50 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
                     fprintf(stderr,"ERR:  Menus do not have chapters\n");
                     return 0;
                 }
-                p=(ismenu?ws->menus->groups[ws->curmenu].pg:ws->titles)->pgcs[thispgc];
                 switch(cs->i3>>16) {
-                case 0: numc=p->numchapters; des="chapter"; break;
-                case 1: numc=p->numprograms; des="program"; break;
-                case 2: numc=p->numcells; des="cell"; break;
+                case 0: numc=curpgc->numchapters; des="chapter"; break;
+                case 1: numc=curpgc->numprograms; des="program"; break;
+                case 2: numc=curpgc->numcells; des="cell"; break;
                 default: numc=0; des="<err>"; break;
                 }
                 if( (cs->i3&65535)>numc ) {
                     fprintf(stderr,"ERR:  Cannot jump to %s %d, only %d exist\n",des,cs->i3&65535,numc);
                     return 0;
                 }
-                write8(buf,0x20,0x05+(cs->i3>>16),0x00,0x00,0x00,0x00,0x00,cs->i3);
+                write8(buf,0x20,0x05+(cs->i3>>16),0x00,0x00,0x00,0x00,0x00,cs->i3); // LinkPTTN pttn, LinkPGCN pgn, or LinkCN cn
                 buf+=8;
-            } else if( cs->i2<128 ) {
+            } else if( i2<128 ) {
                 //	VTSM	NONE	MPGC	NOCH
                 //	VMGM	NONE	MPGC	NOCH
                 //	VTSM	NONE	MEPGC	NOCH
                 //	VMGM	NONE	MEPGC	NOCH
-                if( cs->i2>=120 && cs->i2<128 ) {
+                if( !curgroup ) {
+                    fprintf(stderr,"ERR:  Cannot jump to menu; none exist\n");
+                    return 0;
+                } else if( i2>=120 && i2<128 ) {
                     int i;
                     
-                    for( i=0; i<ws->menus->groups[ws->curmenu].pg->numpgcs; i++ )
-                        if( ws->menus->groups[ws->curmenu].pg->pgcs[i]->entries&(1<<(cs->i2-120)) ) {
-                            cs->i2=i+1;
+                    for( i=0; i<curgroup->numpgcs; i++ )
+                        if( curgroup->pgcs[i]->entries&(1<<(i2-120)) ) {
+                            i2=i+1;
                             break;
                         }
-                    if( cs->i2>=120 ) {
-                        fprintf(stderr,"ERR:  Cannot find PGC with entry %s\n",entries[cs->i2-120]);
+                    if( i2>=120 ) {
+                        fprintf(stderr,"ERR:  Cannot find PGC with entry %s\n",entries[i2-120]);
                         return 0;
                     }
                 } else {
-                    if( cs->i2>ws->menus->groups[ws->curmenu].pg->numpgcs ) {
-                        fprintf(stderr,"ERR:  Cannot jump to menu PGC #%d, only %d exist\n",cs->i2,ws->menus->groups[ws->curmenu].pg->numpgcs);
+                    if( i2>curgroup->numpgcs ) {
+                        fprintf(stderr,"ERR:  Cannot jump to menu PGC #%d, only %d exist\n",i2,curgroup->numpgcs);
                         return 0;
                     }
                 }
                 if( ismenu==2 ) {
                     // In case we are jumping from a FP to VMGM, we need to use a JumpSS
                     // instruction
-                    write8(buf,0x30,0x06,0x00,cs->i2&127,0x00,0xc0,0x00,0x00); // JumpSS VMGM pgcn
+                    write8(buf,0x30,0x06,0x00,i2&127,0x00,0xc0,0x00,0x00); // JumpSS VMGM pgcn
                 } else
-                    write8(buf,0x20,0x04,0x00,0x00,0x00,0x00,0x00,cs->i2&127); // LinkPGCN
+                    write8(buf,0x20,0x04,0x00,0x00,0x00,0x00,0x00,i2&127); // LinkPGCN pgcn
                 buf+=8;
             } else {
                 //	VMGM	NONE	TPGC	NOCH
@@ -519,24 +631,29 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
                 //	VTS	NONE	TPGC	CHXX
                 //	VTSM	NONE	TPGC	CHXX
                 if( ismenu<2 ) {
-                    if( cs->i2-128>ws->titles->numpgcs ) {
-                        fprintf(stderr,"ERR:  Cannot jump to title #%d, only %d exist\n",cs->i2-128,ws->titles->numpgcs);
+                    if( i2-128>ws->titles->numpgcs ) {
+                        fprintf(stderr,"ERR:  Cannot jump to title #%d, only %d exist\n",i2-128,ws->titles->numpgcs);
                         return 0;
                     }
-                    if( cs->i3 && cs->i3>ws->titles->pgcs[cs->i2-128-1]->numchapters ) {
-                        fprintf(stderr,"ERR:  Cannot jump to chapter %d of title %d, only %d exist\n",cs->i3,cs->i2-128,ws->titles->pgcs[cs->i2-128-1]->numchapters);
+                    if( cs->i3 && cs->i3>ws->titles->pgcs[i2-128-1]->numchapters ) {
+                        fprintf(stderr,"ERR:  Cannot jump to chapter %d of title %d, only %d exist\n",cs->i3,i2-128,ws->titles->pgcs[i2-128-1]->numchapters);
                         return 0;
                     }
                 }
-                write8(buf,0x30,ismenu==2?0x02:(cs->i3?0x05:0x03),0x00,cs->i3,0x00,cs->i2-128,0x00,0x00);
+                write8(buf,0x30,ismenu==2?0x02:(cs->i3?0x05:0x03),0x00,cs->i3,0x00,i2-128,0x00,0x00);
                 buf+=8;
             }
             break;
+        }
 
         case VM_CALL:
+        {
+            int i2=cs->i2;
+            int i4=cs->i4;
+
             // CALL's from <post> MUST have a resume cell
-            if( !cs->i4 )
-                cs->i4=1;
+            if( !i4 )
+                i4=1;
             if( ismenu ) {
                 //	VTSM	NONE	NOPGC	NOCH
                 //	VMGM	NONE	NOPGC	NOCH
@@ -589,7 +706,7 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
                 fprintf(stderr,"ERR:  Cannot 'call' a menu from another menu, use 'jump' instead\n");
                 return 0;
             }
-            if( cs->i2==0 || cs->i2>=128 ) {
+            if( i2==0 || i2>=128 ) {
                 //	VTS	NONE	NOPGC	NOCH
                 //	VTS	VMGM	NOPGC	NOCH
                 //	VTS	TS	NOPGC	NOCH
@@ -616,31 +733,29 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
                 fprintf(stderr,"ERR:  Cannot 'call' a chatper within a menu\n");
                 return 0;
             }
-            if( cs->i2==121 && (cs->i1>=2 || (cs->i1==0 && ismenu!=2)) ) {
+            if( i2==121 && (cs->i1>=2 || (cs->i1==0 && ismenu!=2)) ) {
                 fprintf(stderr,"ERR:  VMGM must be specified with FPC\n");
                 return 0;
             }
 
 
 
-
-
             if( cs->i1>=2 ) {
                 //	VTS	TS	MPGC	NOCH
                 //	VTS	TS	MEPGC	NOCH
                 if( jumppad ) {
-                    write8(buf,0x71,0x00,0x00,0x0F,cs->i2,cs->i1,0x00,0x00); buf+=8;
+                    write8(buf,0x71,0x00,0x00,0x0F,i2,cs->i1,0x00,0x00); buf+=8;
                     write8(buf,0x71,0x00,0x00,0x0E,0x00,cs->i3,0x00,0x00); buf+=8;
-                    write8(buf,0x30,0x08,0x00,0x00,cs->i4,0x42,0x00,0x00); buf+=8;
+                    write8(buf,0x30,0x08,0x00,0x00,i4,0x42,0x00,0x00); buf+=8;
                 } else {
                     fprintf(stderr,"ERR:  Cannot call to a menu in another titleset\n");
                     return 0;
                 }
-            } else if( cs->i1==0 && cs->i2<120 ) {
+            } else if( cs->i1==0 && i2<120 ) {
                 //	VTS	NONE	MPGC	NOCH
                 if( jumppad ) {
-                    write8(buf,0x71,0x00,0x00,0x0F,cs->i2,0x00,0x00,0x00); buf+=8;
-                    write8(buf,0x30,0x08,0x00,0x00,cs->i4,0x87,0x00,0x00); buf+=8;
+                    write8(buf,0x71,0x00,0x00,0x0F,i2,0x00,0x00,0x00); buf+=8;
+                    write8(buf,0x30,0x08,0x00,0x00,i4,0x87,0x00,0x00); buf+=8;
                 } else {
                     fprintf(stderr,"ERR:  Cannot call to a specific menu PGC, only an entry\n");
                     return 0;
@@ -649,33 +764,34 @@ static unsigned char *compilecs(unsigned char *obuf,unsigned char *buf,struct wo
                 //	VTS	VMGM	MPGC	NOCH
                 //	VTS	VMGM	MEPGC	NOCH
                 // we cannot provide error checking when jumping to a VMGM
-                if( cs->i2==120 )
-                    cs->i2=122;
-                if( cs->i2<120 )
-                    write8(buf,0x30,0x08,0x00,cs->i2,cs->i4,0xC0,0x00,0x00);
+                if( i2==120 )
+                    i2=122;
+                if( i2<120 )
+                    write8(buf,0x30,0x08,0x00,i2,i4,0xC0,0x00,0x00);
                 else
-                    write8(buf,0x30,0x08,0x00,0x00,cs->i4,cs->i2==121?0:(0x40+cs->i2-120),0x00,0x00);
+                    write8(buf,0x30,0x08,0x00,0x00,i4,i2==121?0:(0x40+i2-120),0x00,0x00);
                 buf+=8;
             } else {
                 int i,j;
 
                 //	VTS	NONE	MEPGC	NOCH
-                if( cs->i2==120 )
-                    cs->i2=127;
+                if( i2==120 )
+                    i2=127;
                     
                 for( j=0; j<ws->menus->numgroups; j++ ) {
                     struct pgcgroup *pg=ws->menus->groups[j].pg;
                     for( i=0; i<pg->numpgcs; i++ )
-                        if( pg->pgcs[i]->entries&(1<<(cs->i2-120)) )
+                        if( pg->pgcs[i]->entries&(1<<(i2-120)) )
                             goto foundpgc;
                 }
-                fprintf(stderr,"ERR:  Cannot find PGC with entry %s\n",entries[cs->i2-120]);
+                fprintf(stderr,"ERR:  Cannot find PGC with entry %s\n",entries[i2-120]);
                 return 0;
             foundpgc:
-                write8(buf,0x30,0x08,0x00,0x00,cs->i4,0x80+cs->i2-120,0x00,0x00);
+                write8(buf,0x30,0x08,0x00,0x00,i4,0x80+i2-120,0x00,0x00);
                 buf+=8;
             }
             break;
+        }
 
         case VM_NOP:
             break;
@@ -804,12 +920,14 @@ static void deleteinstruction(unsigned char *obuf,unsigned char *buf,unsigned ch
             b2[7]--;
     memmove(b,b+8,*end-(b+8));
     *end-=8;
+    memset(*end,0,8); // clean up tracks (so pgc structure is not polluted)
 }
 
-static int vm_optimize(unsigned char *obuf,unsigned char *buf,unsigned char **end)
+void vm_optimize(unsigned char *obuf,unsigned char *buf,unsigned char **end)
 {
     unsigned char *b;
 
+ again:
     for( b=buf; b<*end; b+=8 ) {
         int curline=(b-obuf)/8+1;
         // if
@@ -826,7 +944,7 @@ static int vm_optimize(unsigned char *obuf,unsigned char *buf,unsigned char **en
             memcpy(b,b+8,8); // move statement
             memset(b+8,0,8); // replace with nop
             applyif(b,ifs);
-            return 1;
+            goto again;
         }
         // if
         // 1. this is a NOP instruction
@@ -837,7 +955,7 @@ static int vm_optimize(unsigned char *obuf,unsigned char *buf,unsigned char **en
             b[4]==0 && b[5]==0 && b[6]==0 && b[7]==0 &&
             (b+8!=*end || countreferences(obuf,buf,*end,curline)==0) ) {
             deleteinstruction(obuf,buf,end,b);
-            return 1;
+            goto again;
         }
         // if
         // 1. the prev instruction is an UNCONDITIONAL jump/goto
@@ -848,7 +966,7 @@ static int vm_optimize(unsigned char *obuf,unsigned char *buf,unsigned char **en
             (b[-8]>>4)<=3 && (b[-7]&0x70)==0 && (b[-7]&15)!=0 &&
             countreferences(obuf,buf,*end,curline)==0 ) {
             deleteinstruction(obuf,buf,end,b);
-            return 1;
+            goto again;
         }
         // if
         // 1. this instruction sets subtitle/angle/audio
@@ -865,7 +983,7 @@ static int vm_optimize(unsigned char *obuf,unsigned char *buf,unsigned char **en
             if( b[8+4] ) b[4]=b[8+4];
             if( b[8+5] ) b[5]=b[8+5];
             deleteinstruction(obuf,buf,end,b+8);
-            return 1;
+            goto again;
         }
         // if
         // 1. this instruction sets the button directly
@@ -883,18 +1001,56 @@ static int vm_optimize(unsigned char *obuf,unsigned char *buf,unsigned char **en
             if( b[8+6]==0 )
                 b[8+6]=b[4];
             deleteinstruction(obuf,buf,end,b);
-            return 1;
+            goto again;
+        }
+        // if
+        // 1. this instruction sets a GPRM/SPRM register
+        // 2. the next instruction is a link command (not NOP)
+        // 3. there are no references to the second instruction
+        // then
+        // combine
+        if( b+8!=*end &&
+            ((b[0]&0xE0)==0x40 || (b[0]&0xE0)==0x60) && (b[1]&0x7f)==0x00 &&
+            b[8]==0x20 && ((b[8+1]&0x7f)==4 || 
+                           (b[8+1]&0x7f)==5 ||
+                           (b[8+1]&0x7f)==6 ||
+                           (b[8+1]&0x7f)==7 ||
+                           ((b[8+1]&0x7f)==1 && (b[8+7]&0x1f)!=0)) &&
+            countreferences(obuf,buf,*end,curline+1)==0 ) {
+            b[1]=b[8+1];
+            b[6]=b[8+6];
+            b[7]=b[8+7];
+            deleteinstruction(obuf,buf,end,b+8);
+            goto again;
         }
     }
-    return 0;
 }
 
-unsigned char *vm_compile(unsigned char *obuf,unsigned char *buf,struct workset *ws,int thispgc,struct vm_statement *cs,int ismenu,int prepost)
+unsigned char *vm_compile(unsigned char *obuf,unsigned char *buf,const struct workset *ws,const struct pgcgroup *curgroup,const struct pgc *curpgc,const struct vm_statement *cs,int ismenu)
 {
-    unsigned char *end=compilecs(obuf,buf,ws,thispgc,cs,ismenu,prepost);
+    unsigned char *end;
+    int i, j;
+
+    numlabels=0;
+    numgotos=0;
+
+    end=compilecs(obuf,buf,ws,curgroup,curpgc,cs,ismenu);
     if( !end ) return end;
 
-    while(vm_optimize(obuf,buf,&end));
+    // fix goto references
+    for( i=0; i<numgotos; i++ ) {
+        for( j=0; j<numlabels; j++ )
+            if( !strcasecmp(gotos[i].lname,labels[j].lname) )
+                break;
+        if( j==numlabels ) {
+            fprintf(stderr,"ERR:  Cannot find label %s\n",gotos[i].lname);
+            return 0;
+        }
+        gotos[i].code[7]=(labels[j].code-obuf)/8+1;
+    }
+
+    vm_optimize(obuf,buf,&end);
+
     return end;
 }
 
@@ -908,8 +1064,8 @@ struct vm_statement *vm_parse(const char *b)
 {
     if( b ) {
         char *cmd=strdup(b);
+        dvdvm_buffer_state buf=dvdvm_scan_string(cmd);
         dvd_vm_parsed_cmd=0;
-        dvdvm_scan_string(cmd);
         if( dvdvmparse() ) {
             fprintf(stderr,"ERR:  Parser failed on code '%s'.\n",b);
             exit(1);
@@ -918,6 +1074,7 @@ struct vm_statement *vm_parse(const char *b)
             fprintf(stderr,"ERR:  Nothing parsed from '%s'\n",b);
             exit(1);
         }
+        dvdvm_delete_buffer(buf);
         free(cmd);
         return dvd_vm_parsed_cmd;
     } else {
