@@ -30,7 +30,6 @@
 
 #include <netinet/in.h>
 
-static const char RCSID[]="$Id: //depot/dvdauthor/src/mpeg2desc.c#33 $";
 
 // #define SHOWDATA
 
@@ -410,7 +409,10 @@ int main(int argc,char **argv)
                 if( buf[4]&64 )
                     forceread(buf+5,2,stdin);
                 if( outputenglish ) {
-                    printf("%08x: picture coding extension hdr%s\n",disppos,(buf[3]&2)?", repeat":"");
+                    printf("%08x: picture coding extension hdr%s%s\n",
+                           disppos,
+                           (buf[3]&0x80)?", top":", bottom",
+                           (buf[3]&2)?", repeat":"");
                 }
                 break;
             default:
@@ -421,6 +423,12 @@ int main(int argc,char **argv)
             forceread(&hdr,4,stdin);
             break;
 
+        case 0x1b7: // end of sequence
+            if( outputenglish )
+                printf("%08x: end of sequence\n",disppos);
+            forceread(&hdr,4,stdin);
+            break;
+
         case 0x1b8: // group of pictures
             forceread(buf,4,stdin);
             if( outputenglish ) {
@@ -446,7 +454,7 @@ int main(int argc,char **argv)
 
         case 0x1ba: // mpeg_pack_header
         {
-            long scr,scrhi,scrext;
+            uint32_t scr,scrhi,scrext;
             int64_t fulltime;
             forceread(buf,8,stdin);
             if((buf[0] & 0xC0) == 0x40) {
@@ -462,7 +470,7 @@ int main(int argc,char **argv)
                 scrext=((buf[4]&3)<<7)|
                     (buf[5]>>1);
                 if( scrext >= 300 && outputenglish ) {
-                    printf("WARN: scrext in pack hdr > 300: %ld\n",scrext);
+                    printf("WARN: scrext in pack hdr > 300: %u\n",scrext);
                 }
                 fulltime=((int64_t)scrhi)<<32|((int64_t)scr);
                 fulltime*=300;
@@ -479,7 +487,7 @@ int main(int argc,char **argv)
             }
 
             if( outputenglish )
-                printf("%08x: mpeg%c pack hdr, %lld.%03lld sec\n",disppos,mpeg2?'2':'1',fulltime/SCRTIME,(fulltime%SCRTIME)/(SCRTIME/1000));
+                printf("%08x: mpeg%c pack hdr, %" PRId64 ".%03" PRId64 " sec\n",disppos,mpeg2?'2':'1',fulltime/SCRTIME,(fulltime%SCRTIME)/(SCRTIME/1000));
             
             forceread(&hdr,4,stdin);
             break;
@@ -617,14 +625,14 @@ int main(int argc,char **argv)
                         
                         pts=readpts(buf+eptr);
                         eptr+=5;
-                        printf("; pts %lld.%03lld sec",pts/PTSTIME,(pts%PTSTIME)/(PTSTIME/1000));
+                        printf("; pts %" PRId64 ".%03" PRId64 " sec",pts/PTSTIME,(pts%PTSTIME)/(PTSTIME/1000));
                     }
                     if( has_dts ) {
                         int64_t dts;
                     
                         dts=readpts(buf+eptr);
                         eptr+=5;
-                        printf("; dts %lld.%03lld sec",dts/PTSTIME,(dts%PTSTIME)/(PTSTIME/1000));
+                        printf("; dts %" PRId64 ".%03" PRId64 " sec",dts/PTSTIME,(dts%PTSTIME)/(PTSTIME/1000));
                     }
 		    if(mpeg2) {
                         if( buf[1] & 32 ) {
