@@ -24,6 +24,7 @@
 #include <assert.h>
 #include <ctype.h>
 #include <errno.h>
+#include <fcntl.h>
 
 #include <libxml/xmlreader.h>
 
@@ -34,20 +35,29 @@
 #include <locale.h>
 #endif
 
-static const char RCSID[]="$Id: //depot/dvdauthor/src/readxml.c#17 $";
 
 int parser_err=0, parser_acceptbody=0;
 char *parser_body=0;
 
+static int xml_varied_read(void *context,char *buffer,int len)
+{
+    return fread(buffer,1,len,((struct vfile *)context)->h);
+}
+
+static int xml_varied_close(void *context)
+{
+    varied_close(*((struct vfile *)context));
+    return 0;
+}
+
 int readxml(const char *xmlfile,struct elemdesc *elems,struct elemattr *attrs)
 {
     int curstate=0,statehistory[10];
     xmlTextReaderPtr f;
+    struct vfile fd;
 
-    if( xmlfile[0]=='&' && isdigit(xmlfile[1]) )
-        f=xmlReaderForFd(atoi(xmlfile+1),xmlfile,NULL,0);
-    else
-        f=xmlNewTextReaderFilename(xmlfile);
+    fd=varied_open(xmlfile,O_RDONLY);
+    f=xmlReaderForIO(xml_varied_read,xml_varied_close,&fd,xmlfile,NULL,0);
     if(!f) {
         fprintf(stderr,"ERR:  Unable to open XML file %s\n",xmlfile);
         return 1;
@@ -71,11 +81,12 @@ int readxml(const char *xmlfile,struct elemdesc *elems,struct elemattr *attrs)
         case XML_READER_TYPE_COMMENT:
             break;
 
-        case XML_READER_TYPE_ELEMENT:
+        case XML_READER_TYPE_ELEMENT: {
+            char *elemname=xmlTextReaderName(f);
             assert(!parser_body);
             for( i=0; elems[i].elemname; i++ )
                 if( curstate==elems[i].parentstate &&
-                    !strcmp(xmlTextReaderName(f),elems[i].elemname) ) {
+                    !strcmp(elemname,elems[i].elemname) ) {
                     // reading the attributes causes these values to change
                     // so if you want to use them later, save them now
                     int empty=xmlTextReaderIsEmptyElement(f),
@@ -104,6 +115,8 @@ int readxml(const char *xmlfile,struct elemdesc *elems,struct elemattr *attrs)
                                     fprintf(stderr,"ERR:      %s\n",attrs[j].attr);
                             return 1;
                         }
+                        xmlFree(nm);
+                        xmlFree(v);
                     }
                     if( empty ) {
                         if( elems[i].end ) {
@@ -118,13 +131,15 @@ int readxml(const char *xmlfile,struct elemdesc *elems,struct elemattr *attrs)
                     break;
                 }
             if( !elems[i].elemname ) {
-                fprintf(stderr,"ERR:  Cannot match start tag '%s'.  Valid tags are:\n",xmlTextReaderName(f));
+                fprintf(stderr,"ERR:  Cannot match start tag '%s'.  Valid tags are:\n",elemname);
                 for( i=0; elems[i].elemname; i++ )
                     if( curstate==elems[i].parentstate )
                         fprintf(stderr,"ERR:      %s\n",elems[i].elemname);
                 return 1;
             }
+            xmlFree(elemname);
             break;
+        }
 
         case XML_READER_TYPE_END_ELEMENT:
             i=statehistory[xmlTextReaderDepth(f)];
@@ -134,6 +149,8 @@ int readxml(const char *xmlfile,struct elemdesc *elems,struct elemattr *attrs)
                     return 1;
             }
             curstate=elems[i].parentstate;
+            if( parser_body )
+                free(parser_body);
             parser_body=0;
             parser_acceptbody=0;
             if( !curstate )
@@ -152,6 +169,7 @@ int readxml(const char *xmlfile,struct elemdesc *elems,struct elemattr *attrs)
                         v[i]!=' '  &&
                         v[i]!='\t' )
                         goto has_nonws_body;
+                xmlFree(v);
                 break;
             }
         has_nonws_body:
@@ -166,6 +184,7 @@ int readxml(const char *xmlfile,struct elemdesc *elems,struct elemattr *attrs)
                 parser_body=realloc(parser_body,strlen(parser_body)+strlen(v)+1);
                 strcat(parser_body,v);
             }
+            xmlFree(v);
             break;
         }
 
@@ -175,6 +194,7 @@ int readxml(const char *xmlfile,struct elemdesc *elems,struct elemattr *attrs)
         }
     }
  done_parsing:
+    xmlFreeTextReader(f);
 
     return 0;
 }
