@@ -44,14 +44,14 @@
 #define CBUFSIZE 65536
 #define PSBUFSIZE 10
 
-static const char RCSID[]="$Id: //depot/dvdauthor/src/spuunmux.c#21 $";
 
 static unsigned int add_offset;
 
 static int debug = 0;
 
 static int full_size = FALSE;
-static unsigned int ofs, ofs1, pts, spts, subi, subs, subno;
+static unsigned int pts, spts, subi, subs, subno;
+static int ofs, ofs1;
 static unsigned char sub[65536];
 static unsigned char next_bits;
 static char *base_name;
@@ -83,9 +83,9 @@ struct button {
 struct dispdetails {
     int pts[2];
     int numpal;
-    u_int32_t palette[16];
+    uint32_t palette[16];
     int numcoli;
-    u_int32_t coli[6];
+    uint32_t coli[6];
     int numbuttons;
     struct button *buttons;
     struct dispdetails *next;
@@ -94,8 +94,8 @@ struct dispdetails {
 static struct dispdetails *dd=0;
 
 struct colormap {
-    u_int16_t color;
-    u_int16_t contrast;
+    uint16_t color;
+    uint16_t contrast;
     int x1,y1,x2,y2;
 };
 
@@ -160,15 +160,17 @@ static void adddd(struct dispdetails *d)
 static int dvddecode()
 {
     unsigned int io;
-    unsigned short int size, dsize, i, x, y, t;
+    uint16_t size, dsize, i, x, y, t;
     unsigned char c;
     struct spu *s;
 
-    size = (((unsigned int) sub[0]) << 8) + sub[1];
-    dsize = (((unsigned int) sub[2]) << 8) + sub[3];
+    size = read2(sub);
+    dsize = read2(sub+2);
+
+    ofs = -1;
 
     if (debug > 1)
-	fprintf(stderr, "packet: 0x%x bytes, 0x%x bytes data\n", size,
+	fprintf(stderr, "packet: %d bytes, first block offset=%d\n", size,
 		dsize);
 
     s=malloc(sizeof(struct spu));
@@ -180,48 +182,41 @@ static int dvddecode()
     s->nummap=1;
     s->map=malloc(sizeof(struct colormap));
     memset(s->map,0,sizeof(struct colormap));
-    s->map[0].x2=0x7ffffff;
-    s->map[0].y2=0x7ffffff;
+    s->map[0].x2=0x7fffffff;
+    s->map[0].y2=0x7fffffff;
     i = dsize + 4;
 
-    t = sub[dsize] * 256 + sub[dsize + 1];
+    t = read2(sub+dsize);
 
     if (debug > 2)
-	fprintf(stderr, "time offset: %d sub[%d]=0x%02x sub[%d]=0x%02x\n",
-		t, dsize, sub[dsize], dsize + 1, sub[dsize + 1]);
+	fprintf(stderr, "\tBLK(%5d): time offset: %d; next: %d\n", dsize, t, read2(sub+dsize+2));
 
     while (i < size) {
 	c = sub[i];
-	if (debug > 4)
-	    fprintf(stderr, "i: %d cmd: 0x%x\n", i, c);
 
 	switch (c) {
 	case 0x0:		//force start display
-	    i++;
+	    if (debug > 4)
+		fprintf(stderr, "\tcmd(%5d): force start display\n",i);
 	    s->force_display = TRUE;
-	    s->pts[0] = t * 1024 + spts;
-	    if (debug > 4)
-		fprintf(stderr, "cmd: force start display\n");
-	    break;
-
+            // fall through
 	case 0x01:
+	    if (debug > 4 && c==0x01)
+		fprintf(stderr, "\tcmd(%5d): start display\n",i);
 	    i++;
 	    s->pts[0] = t * 1024 + spts;
-	    if (debug > 4)
-		fprintf(stderr, "cmd: start display\n");
 	    break;
 
 	case 0x02:
 	    if (debug > 4)
-		fprintf(stderr, "cmd: end display\n");
+		fprintf(stderr, "\tcmd(%5d): end display\n",i);
 	    s->pts[1] = t * 1024 + spts;
 	    i++;
 	    break;
 
 	case 0x03:
 	    if (debug > 4)
-		fprintf(stderr, "cmd: palette %02x%02x\n", sub[i + 1],
-			sub[i + 2]);
+		fprintf(stderr, "\tcmd(%5d): palette=%02x%02x\n", i, sub[i + 1], sub[i + 2]);
 
             s->map[0].color=read2(sub+i+1);
 	    i += 3;
@@ -229,8 +224,7 @@ static int dvddecode()
 
 	case 0x04:
 	    if (debug > 4)
-		fprintf(stderr, "cmd: t-palette %02x%02x\n", sub[i + 1],
-			sub[i + 2]);
+		fprintf(stderr, "\tcmd(%5d): transparency=%02x%02x\n", i, sub[i + 1], sub[i + 2]);
 
             s->map[0].contrast=read2(sub+i+1);
 	    i += 3;
@@ -244,39 +238,36 @@ static int dvddecode()
 	    s->yd = (((sub[i + 5] & 0x0f) << 8) + sub[i + 6]) - s->y0 + 1;
 
 	    if (debug > 4)
-		fprintf(stderr, "cmd: img ofs %d,%d  size: %d,%d\n", s->x0,
+		fprintf(stderr, "\tcmd(%5d): image corner=%d,%d, size=%d,%d\n", i, s->x0,
 			s->y0, s->xd, s->yd);
 	    i += 7;
 	    break;
 
 	case 0x06:
-	    ofs = (((unsigned int) sub[i + 1]) << 8) + sub[i + 2];
-	    ofs1 = (((unsigned int) sub[i + 3]) << 8) + sub[i + 4];
+            if( ofs>=0 )
+                fprintf(stderr,"WARN: image pointer already supplied for this subpicture\n");
+	    ofs = read2(sub+i+1);
+	    ofs1 = read2(sub+i+3);
 	    if (debug > 4)
-		fprintf(stderr, "cmd: image offsets 0x%x 0x%x\n", ofs,
+		fprintf(stderr, "\tcmd(%5d): image offsets=%d,%d\n", i, ofs,
 			ofs1);
 	    i += 5;
 	    break;
 
 	case 0xff:
-	    if (debug > 4)
-		fprintf(stderr, "cmd: end cmd\n");
-
 	    if (i + 5 > size) {
 		if (debug > 4)
-		    fprintf(stderr,
-			    "short end command i=%d size=%d (%d bytes)\n",
-			    i, size, size - i);
+		    fprintf(stderr,"\tcmd(%5d): end cmd\n",i);
 
 		i = size;
 		break;
 	    }
 
-	    t = sub[i + 1] * 256 + sub[i + 2];
+	    t = read2(sub + i + 1);
 	    if (debug > 4) {
-		fprintf(stderr, "next packet time: %d  end: 0x%02x%02x\n",
-			t, sub[i + 3], sub[i + 4]);
-	    }
+		fprintf(stderr, "\tcmd(%5d): end cmd\n",i);
+                fprintf(stderr, "\tBLK(%5d): time offset: %d; next: %d\n", i+1, t, read2(sub+i+3));
+            }
 
 	    if ((sub[i + 3] != sub[dsize + 2])
 		|| (sub[i + 4] != sub[dsize + 3])) {
@@ -287,9 +278,6 @@ static int dvddecode()
 			    sub[dsize + 3], dsize);
 		}
 
-		if (debug > 3)
-		    fprintf(stderr, "i: %d\n", i);
-
 		i = size;
 		break;
 	    }
@@ -298,6 +286,8 @@ static int dvddecode()
 	    break;
 
 	default:
+            if (debug > 4)
+                fprintf(stderr, "\tcmd(%5d): 0x%x\n", i, c);
 	    if (debug > 0)
 		fprintf(stderr,
 			"invalid sequence in control header (%02x)!\n", c);
@@ -310,50 +300,42 @@ static int dvddecode()
     io = 0;
     s->img = malloc( s->xd*s->yd);
 
-    while ((ofs < dsize) && (y < s->yd)) {
-	i = get_next_bits();
+    if( ofs<0 && y<s->yd ) {
+        fprintf(stderr,"WARN: No image data supplied for this subtitle\n");
+    } else {
+        while ((ofs < dsize) && (y < s->yd)) {
+            i = get_next_bits();
 
-	if (i < 4) {
-	    i = (i << 4) + get_next_bits();
-	    if (i < 16) {
-		i = (i << 4) + get_next_bits();
-		if (i < 0x40) {
-		    i = (i << 4) + get_next_bits();
-		    if (i < 256) {
-			have_bits = FALSE;
+            if (i < 4) {
+                i = (i << 4) + get_next_bits();
+                if (i < 16) {
+                    i = (i << 4) + get_next_bits();
+                    if (i < 0x40) {
+                        i = (i << 4) + get_next_bits();
+                        if (i < 4) {
+                            i=i+(s->xd-x)*4;
+                        }
+                    }
+                }
+            }
 
-			y += 2;
-			while (x++ != s->xd)
-			    s->img[io++] = i;
-			x = 0;
-			if ((y >= s->yd) && !(y & 1)) {
-			    y = 1;
-			    io = s->xd;
-			    ofs = ofs1;
-			} else
-			    io += s->xd;
-			continue;
-		    }
-		}
-	    }
-	}
-
-	c = i & 3;
-	i = i >> 2;
-	while (i--) {
-	    s->img[io++] = c;
-	    if (++x == s->xd) {
-		y += 2;
-		x = 0;
-                if ((y >= s->yd) && !(y & 1)) {
-                    y = 1;
-                    io = s->xd;
-                    ofs = ofs1;
-                } else
-                    io += s->xd;
-		have_bits = FALSE;
-	    }
-	}
+            c = i & 3;
+            i = i >> 2;
+            while (i--) {
+                s->img[io++] = c;
+                if (++x == s->xd) {
+                    y += 2;
+                    x = 0;
+                    if ((y >= s->yd) && !(y & 1)) {
+                        y = 1;
+                        io = s->xd;
+                        ofs = ofs1;
+                    } else
+                        io += s->xd;
+                    have_bits = FALSE;
+                }
+            }
+        }
     }
     
     if (s->pts[0] == -1)
@@ -365,9 +347,6 @@ static int dvddecode()
     
     addspu(s);
 
-    if (debug > 2)
-	fprintf(stderr, "ofs: 0x%x y: %d\n", ofs, y);
-
     return 0;
 }				/* end fuction dvd_decode */
 
@@ -571,16 +550,52 @@ static void write_pts(char *preamble,int pts)
             (pts / 900) % 100);
 }
 
+/*
+  copy the content of buf to expbuf converting '&' '<' '>' '"'
+  expbuf must be big enough to contain the expanded buffer
+*/
+static void xml_buf(unsigned char *expbuf,unsigned char *buf)
+{
+    unsigned char *p;
+
+    do {
+        switch (*buf) {
+            case '&':
+              p="&amp;";
+              break;
+            case '<':
+              p="&lt;";
+              break;
+            case '>':
+              p="&gt;";
+              break;
+            case '"':
+              p="&quot;";
+              break;
+            default:
+              p=NULL;
+              break;
+        }
+        if( p ) {
+            while ( (*expbuf++ = *p++) )
+                 ;
+            --expbuf;
+        } else
+            *expbuf++ = *buf;
+    } while ( *buf++ ) ;
+}
+
 static void write_menu_image(struct spu *s,struct dispdetails *d,char *type,int offset)
 {
     unsigned char nbuf[256];
+    unsigned char ebuf[256*6];
     int nummap=d->numbuttons+1, i;
     struct colormap *map=malloc(sizeof(struct colormap)*nummap);
     memset(map,0,sizeof(struct colormap)); // set the first one blank
     map[0].x2=0x7fffffff;
     map[0].y2=0x7fffffff;
     for( i=0; i<d->numbuttons; i++ ) {
-        u_int32_t cc=d->coli[2*d->buttons[i].grp-2+offset];
+        uint32_t cc=d->coli[2*d->buttons[i].grp-2+offset];
         map[i+1].x1=d->buttons[i].x1;
         map[i+1].y1=d->buttons[i].y1;
         map[i+1].x2=d->buttons[i].x2;
@@ -590,14 +605,17 @@ static void write_menu_image(struct spu *s,struct dispdetails *d,char *type,int
     }
     
     sprintf(nbuf, "%s%05d%c.png", base_name, s->subno, type[0]);
-    if( !write_png(nbuf,s,map,nummap) )
-        fprintf(fdo," %s=\"%s\"",type,nbuf);
+    if( !write_png(nbuf,s,map,nummap) ) {
+        xml_buf(ebuf,nbuf);
+        fprintf(fdo," %s=\"%s\"",type,ebuf);
+    }
     free(map);
 }
 
 static void write_spu(struct spu *s,struct dispdetails *d)
 {
     unsigned char nbuf[256];
+    unsigned char ebuf[256*6];
     int i;
 
     if( d )
@@ -605,8 +623,10 @@ static void write_spu(struct spu *s,struct dispdetails *d)
     fprintf(fdo,"\t\t<spu");
 
     sprintf(nbuf, "%s%05d.png", base_name, s->subno);
-    if( !write_png(nbuf,s,s->map,s->nummap) )
-        fprintf(fdo," image=\"%s\"",nbuf);
+    if( !write_png(nbuf,s,s->map,s->nummap) ) {
+        xml_buf(ebuf,nbuf);
+        fprintf(fdo," image=\"%s\"",ebuf);
+    }
 
     if( d && d->numbuttons ) {
         write_menu_image(s,d,"highlight",0);
@@ -625,9 +645,9 @@ static void write_spu(struct spu *s,struct dispdetails *d)
         for( i=0; i<d->numbuttons; i++ ) {
             struct button *b=d->buttons+i;
             if( b->autoaction ) 
-                fprintf(fdo,"\t\t\t<action label=\"%s\" />\n",b->name);
+                fprintf(fdo,"\t\t\t<action name=\"%s\" />\n",b->name);
             else {
-                fprintf(fdo,"\t\t\t<button label=\"%s\" x0=\"%d\" y0=\"%d\" x1=\"%d\" y1=\"%d\" up=\"%s\" down=\"%s\" left=\"%s\" right=\"%s\" />\n",
+                fprintf(fdo,"\t\t\t<button name=\"%s\" x0=\"%d\" y0=\"%d\" x1=\"%d\" y1=\"%d\" up=\"%s\" down=\"%s\" left=\"%s\" right=\"%s\" />\n",
                         b->name,b->x1,b->y1,b->x2,b->y2,
                         b->up,b->down,b->left,b->right);
             }
@@ -689,7 +709,7 @@ static void usage(void)
 
 int main(int argc, char **argv)
 {
-    int fd;
+    struct vfile fd;
     int option, n;
     int rgb;
     char *temp;
@@ -808,8 +828,8 @@ int main(int argc, char **argv)
     add_offset = 450; // for rounding purposes
 
     while (inc < Inc) {
-	fd = open(iname[inc], O_RDONLY | O_BINARY);
-	if (fd < 0) {
+	fd = varied_open(iname[inc], O_RDONLY);
+	if (fd.h == 0) {
 	    fprintf(stderr, "error opening file %s\n", iname[inc]);
 
 	    exit(-1);
@@ -820,7 +840,7 @@ int main(int argc, char **argv)
 
 	inc++;
 
-	while (read(fd, &c, 4) == 4) {
+	while (fread(&c, 1, 4, fd.h) == 4) {
             c=ntohl(c);
 	    if (c == 0x000001ba) {	// start PS (Program stream)
 		static unsigned int old_system_time = -1;
@@ -829,15 +849,21 @@ int main(int argc, char **argv)
 		if (debug > 5)
 		    fprintf(stderr, "pack_start_code\n");
 
-		if (read(fd, psbuf, PSBUFSIZE) < 1)
+		if (fread(psbuf, 1, PSBUFSIZE, fd.h) < 1)
 		    break;
 
+		if ( (psbuf[0] & 0xc0) != 0x40 ) {
+		    if (debug > 1)
+		        fprintf(stderr, "not a MPEG-2 file, skipping.\n");
+		    break;
+		}
+
 		new_system_time = (psbuf[4] >> 3) + (psbuf[3] * 32) +
 		    ((psbuf[2] & 3) * 32 * 256) +
 		    ((psbuf[2] & 0xf8) * 32 * 128) +
 		    (psbuf[1] * 1024 * 1024) +
 		    ((psbuf[0] & 3) * 1024 * 1024 * 256) +
-		    ((psbuf[0] & 0x38) * 1024 * 1024 * 256 * 2);
+		    ((psbuf[0] & 0x38) * 1024 * 1024 * 128);
 
 		if (new_system_time < old_system_time) {
 		    if (old_system_time != -1) {
@@ -852,13 +878,23 @@ int main(int argc, char **argv)
                 flushspus(old_system_time);
 
 		if (debug > 5) {
-		    fprintf(stderr, "system time: %d\n", new_system_time);
+		    fprintf(stderr, "system time: %u\n", new_system_time);
+		}
+
+                c=psbuf[9]&7;
+		if (c) {
+                    char s[7];
+
+		    if (debug > 5)
+		        fprintf(stderr, "found %d stuffing bytes\n", c);
+		    if (fread(s, 1, c, fd.h) < c)
+		        break;
 		}
             } else if( c==0x1b9 ) {
                 if (debug > 5)
                     fprintf(stderr, "end packet\n");
 	    } else {
-		read(fd, &package_length, 2);
+		fread(&package_length, 1, 2, fd.h);
                 package_length=ntohs(package_length);
 		if (package_length != 0) {
 
@@ -874,7 +910,7 @@ int main(int argc, char **argv)
 		    case 0x01bd:
 			if (debug > 5)
 			    fprintf(stderr, "private stream\n");
-			read(fd, cbuf, package_length);
+			fread(cbuf, 1, package_length, fd.h);
 
 			next_word = getpts(cbuf);
 			if (next_word != -1) {
@@ -945,7 +981,7 @@ int main(int argc, char **argv)
 			break;
 		    case 0x01e0:
                         if( firstvideo==-1 ) {
-                            read(fd, cbuf, package_length);
+                            fread(cbuf, 1, package_length, fd.h);
                             firstvideo=getpts(cbuf);
                             add_offset-=firstvideo;
                             package_length=0;
@@ -975,7 +1011,7 @@ int main(int argc, char **argv)
 			if (debug > 5)
 			    fprintf(stderr, "padding stream %d bytes\n",
 				    package_length);
-			read(fd, cbuf, package_length);
+			fread(cbuf, 1, package_length, fd.h);
                         if( package_length > 30 ) {
                             int i;
 
@@ -1005,7 +1041,7 @@ int main(int argc, char **argv)
 
                                         d->numpal=0;
                                         for( j=0; j<cbuf[i+1]; j++ ) {
-                                            int c=(cbuf[i+2+3*j]<<16)|(cbuf[i+3+3*j]<<8)|(cbuf[i+4+3*j]);
+                                            int c=read4(cbuf+i+1+3*j)&0xffFFff;
                                             d->palette[j]=c;
                                             d->numpal++;
                                         }
@@ -1027,6 +1063,7 @@ int main(int argc, char **argv)
                                         int j;
                                         d->numbuttons=cbuf[i+1];
                                         d->buttons=malloc(d->numbuttons*sizeof(struct button));
+                                        i+=2;
                                         for( j=0; j<d->numbuttons; j++ ) {
                                             struct button *b=&d->buttons[j];
                                             b->name=readpstr(cbuf,&i);
@@ -1103,7 +1140,7 @@ int main(int argc, char **argv)
 			package_length = 2;
 			while (next_word != 0x1ba) {
 			    next_word = next_word << 8;
-			    if (read(fd, &next_word, 1) < 1)
+			    if (fread(&next_word, 1, 1, fd.h) < 1)
 				break;
 			    package_length++;
 			}
@@ -1114,14 +1151,14 @@ int main(int argc, char **argv)
 				    package_length);
 			goto l_01ba;
 		    }		/* end switch */
-                    read(fd, cbuf, package_length);
+                    fread(cbuf, 1, package_length, fd.h);
 		}
 
 	    }			/* end if 0xbd010000 */
 
 	}			/* end while read 4 */
 
-	close(fd);
+	varied_close(fd);
     }				/* end while inc < Inc */
 
     flushspus(0x7fffffff);
