@@ -26,6 +26,7 @@
 #include <math.h>
 
 #ifdef HAVE_MAGICK
+#include <stdarg.h>
 #include <magick/api.h>
 #else
 #include <png.h>
@@ -34,7 +35,6 @@
 #include "subgen.h"
 #include "textsub.h"
 
-static const char RCSID[]="$Id: //depot/dvdauthor/src/subgen-image.c#23 $";
 
 #define MAXX 720
 #define MAXY 576
@@ -312,6 +312,7 @@ static int read_pic(stinfo *s,pict *p)
     if ( have_textsub)
     {
         textsub_render(s->sub_title);
+        s->forced = s->sub_title->text_forced;
         r=read_frame(p);
     }
     else
@@ -361,12 +362,26 @@ static int pickbuttongroups(stinfo *s,int ng,int useimg)
     gs=malloc(ng*sizeof(palgroup));
     memset(gs,0,ng*sizeof(palgroup));
 
+    assert(!useimg || (s->xd <= s->img.width && s->yd <= s->img.height));
+    assert(s->xd <= s->hlt.width && s->yd <= s->hlt.height);
+    assert(s->xd <= s->sel.width && s->yd <= s->sel.height);
+
     // fprintf(stderr,"attempt %d groups, %d useimg\n",ng,useimg);
     // find unique colors per button
     for( i=0; i<s->numbuttons; i++ ) {
         button *b=&s->buttons[i];
         palgroup *bp=&bpgs[i];
 
+        if( b->r.x0 != b->r.x1 && b->r.y0 != b->r.y1
+	    && (b->r.x0 < 0 || b->r.x0 > b->r.x1 || b->r.x1 > s->xd
+		|| b->r.y0 < 0 || b->r.y0 > b->r.y1 || b->r.y1 > s->yd) )
+        {
+            if (debug > -1)
+                fprintf(stderr,
+                        "ERR: Button coordinates out of range: (%d,%d)-(%d,%d)\n",
+                        b->r.x0, b->r.y0, b->r.x1, b->r.y1);
+            exit(1);
+        }
         for( y=b->r.y0; y<b->r.y1; y++ )
             for( x=b->r.x0; x<b->r.x1; x++ )
                 if( !checkcolor(bp,gettricolor(s,y*s->xd+x,useimg)) )
@@ -438,6 +453,50 @@ static int pickbuttongroups(stinfo *s,int ng,int useimg)
         }
         free(bpgs);
         free(gs);
+
+	// If possible, make each palette entry 0 transparent in
+	// all states, since some players may pad buttons with 0
+	// and we also pad with 0 in some cases.
+	for( j=0; j<ng; j++ ) {
+	    int spare = -1;
+	    int tri;
+
+            // search for an unused color, or one that is already fully transparent
+	    for( k=0; k<4; k++ ) {
+		tri = s->groupmap[j][k];
+		if( tri == -1
+		    || (s->img.pal[(tri >> 16) & 0xFF].t == 0
+			&& s->hlt.pal[(tri >> 8) & 0xFF].t == 0
+			&& s->sel.pal[tri & 0xFF].t == 0) ) {
+		    spare = k;
+		    break;
+		}
+	    }
+            // if we find a spare color but one that is not transparent, make one that is transparent
+	    if( spare > 0 && tri == -1 ) {
+		tri = 0;
+		for( k=0; k<s->img.numpal; ++k )
+		    if( s->img.pal[k].t == 0 ) {
+			tri |= k << 16;
+			break;
+		    }
+		for( k=0; k<s->hlt.numpal; ++k )
+		    if( s->hlt.pal[k].t == 0 ) {
+			tri |= k << 8;
+			break;
+		    }
+		for( k=0; k<s->sel.numpal; ++k )
+		    if( s->sel.pal[k].t == 0 ) {
+			tri |= k;
+			break;
+		    }
+	    }
+	    if( spare > 0 ) {
+		s->groupmap[j][spare] = s->groupmap[j][0];
+		s->groupmap[j][0] = tri;
+	    }
+	}
+
         fprintf(stderr,"INFO: Pickbuttongroups, success with %d groups, useimg=%d\n",ng,useimg);
         s->numgroups=ng;
         return 1;
@@ -641,6 +700,9 @@ static void detectbuttons(stinfo *s)
                     }
                 } while(didwork);
 
+                r.y0-=r.y0&1; // buttons need even 'y' coordinates
+                r.y1+=r.y1&1;
+
                 // add button r
                 rs=realloc(rs,(numr+1)*sizeof(rectangle));
                 rs[numr++]=r;
