@@ -35,7 +35,6 @@
 #include "subgen.h"
 #include "textsub.h"
 
-static const char RCSID[]="$Id: //depot/dvdauthor/src/subgen.c#49 $";
 
 
 // (90000*300)/(1260000/2048)
@@ -62,7 +61,7 @@ static unsigned int spuindex, progr;
 static int tofs;
 static int svcd;
 
-static u_int64_t lps;
+static uint64_t lps;
 
 stinfo **spus=0;
 int numspus=0;
@@ -84,7 +83,7 @@ int max_sub_size;
 static int subno,secsize,mode,fdo,header_size,muxrate;
 static unsigned char substr, *sector;
 static stinfo *newsti;
-static u_int64_t gts, nextgts;
+static uint64_t gts, nextgts;
 
 
 /*
@@ -135,7 +134,7 @@ per button:
       db up, down, left, right
 */
 
-static void mkpackh(u_int64_t time, unsigned int muxrate, unsigned char stuffing)
+static void mkpackh(uint64_t time, unsigned int muxrate, unsigned char stuffing)
 {
     unsigned long th=time/300,tl=time%300;
     header[0] = 0x44 | ((th >> 27) & 0x38) | ((th >> 28) & 3);
@@ -191,16 +190,16 @@ static unsigned int getmuxr(unsigned char *buf)
     return (buf[8] >> 2)|(buf[7]*64)|(buf[6]*16384);
 }
 
-static u_int64_t getgts(unsigned char *buf)
+static uint64_t getgts(unsigned char *buf)
 {
-    u_int64_t th,tl;
+    uint64_t th,tl;
     if (((buf[8]&3) != 3) || ((buf[5]&1) != 1) || ((buf[4]&4) != 4)||((buf[2]&4) != 4)|| ((buf[0]&0xc4) != 0x44)) return -1;
-    th=(buf[4] >> 3) + (buf[3]*32) + ((buf[2]&3)*32*256) + ((buf[2]&0xf8)*32*128) + (buf[1]*1024*1024) + ((buf[0]&3)*1024*1024*256) + ((buf[0]&0x38)*1024*1024*256*2);
+    th=(buf[4] >> 3) + (buf[3]*32) + ((buf[2]&3)*32*256) + ((buf[2]&0xf8)*32*128) + (buf[1]*1024*1024) + ((buf[0]&3)*1024*1024*256) + ((buf[0]&0x38)*1024*1024*128);
     tl=((buf[4]&3)<<7)|(buf[5]>>1);
     return th*300+tl;
 }
 
-static void fixgts(u_int64_t *gts,u_int64_t *nextgts)
+static void fixgts(uint64_t *gts,uint64_t *nextgts)
 {
     if( gts[0] < nextgts[0] )
         gts[0]=nextgts[0];
@@ -307,7 +306,8 @@ static int sread(int h,void *b,int l)
             return -1;
         }
         if( !r ) {
-            fprintf(stderr,"WARN:  Read %d, expected %d\n",tr+r,tr+l);
+            if( tr )
+                fprintf(stderr,"WARN:  Read %d, expected %d\n",tr+r,tr+l);
             return tr;
         }
         l-=r;
@@ -377,10 +377,12 @@ static void mux(int eoinput)
         int bytes_send, sub_size;
         unsigned char seq;
         unsigned int q;
-        u_int64_t dgts;
+        int64_t dgts;
         
         /* wait for correct time to insert sub, leave time for vpts to occur */
         dgts=(newsti->spts-.15*90000)*300;
+        if( dgts < 0 )
+            dgts=0;
         if( dgts > gts && !eoinput)
             break;
 
@@ -557,6 +559,8 @@ static void mux(int eoinput)
                         wdshort(b->r.y0);
                         wdshort(b->r.x1);
                         wdshort(b->r.y1);
+                        if( (b->r.y0&1) || (b->r.y1&1) )
+                            fprintf(stderr,"WARN: Button y coordinates are odd for button %s: %dx%d-%dx%d; they may not display properly.\n",b->name,b->r.x0,b->r.y0,b->r.x1,b->r.y1);
                         sprintf(nm1,"%d",i?i:(cursti->numbuttons));
                         sprintf(nm2,"%d",(i+1!=cursti->numbuttons)?(i+2):1);
                         // fprintf(stderr,"BUTTON NAVIGATION FOR %s: up=%s down=%s left=%s right=%s (%s %s)\n",b->name,b->up,b->down,b->left,b->right,nm1,nm2);
@@ -639,15 +643,14 @@ static void mux(int eoinput)
             if(svcd)
             {
                 /* 4 byte svcd header */
-                unsigned char cc = subno >> 8;
+                uint16_t cc = htons(subno);
 
                 swrite(fdo, &substr, 1); // current subtitle stream
 
                 if (bytes_send + i == sub_size) seq |= 128;
                 swrite(fdo, &seq, 1); // packet number in current sub
                 // 0 - up, last packet has bit 7 set
-                swrite(fdo, &cc, 1);    // h current sub nr
-                swrite(fdo, &subno, 1); // l current sub nr
+                swrite(fdo, &cc, 2);
             }
 
             seq++;
@@ -713,7 +716,10 @@ int main(int argc,char **argv)
     }
 //fprintf(stderr, "malloc sub=%p\n", sub);
 
-    cbuf = malloc(65536);
+    if ( !(cbuf = malloc(65536)) ) {
+	fprintf(stderr, "ERR: Could not allocate space for sub buffer, aborting.\n");
+	exit(1);
+    }
 
     image_init();
 
@@ -806,7 +812,10 @@ int main(int argc,char **argv)
 
     skip=0;
 
-    sector=malloc(secsize);
+    if ( !(sector=malloc(secsize)) ) {
+	fprintf(stderr, "ERR: Could not allocate space for sector buffer, aborting.\n");
+	exit(1);
+    }
 
     newsti=getnextsub();
     max_sub_size = 0;
@@ -831,7 +840,7 @@ int main(int argc,char **argv)
             if(progr)
             {
                 if (lps % 1024 * 1024 * 10 < secsize)
-                    fprintf(stderr, "INFO: %lld bytes of data written\r", lps);
+                    fprintf(stderr, "INFO: %" PRIu64 " bytes of data written\r", lps);
             }
 
             if(debug > 5) fprintf(stderr, "INFO: pack_start_code\n");
